.white {
  background-color: white;
}
.success {
  color: gray;
  font-size: 12px;
  height: 24px;
}
.html-container:has(.no-padding) {
  padding: 0;
  height: 24px;
}
.progress-bar {
  background-color: #2b9a66;
}
.progress-level-inner {
  display: none;
}
